@@ -18,6 +18,7 @@ use std::{
 };
 use libc::{
     c_char,
+    c_void,
     size_t,
 };
 
@@ -52,6 +53,7 @@ use crate::{
     RnpContext,
     RnpOutput,
     Keygrip,
+    RnpSignature,
     RnpUserID,
     RnpPasswordFor,
     str_to_rnp_buffer,
@@ -114,6 +116,7 @@ impl RnpKey {
     pub fn new(ctx: *mut RnpContext,
                key: Key<UnspecifiedParts, UnspecifiedRole>,
                cert: &Cert) -> Self {
+        rnp_function!(RnpKey::new, crate::TRACE);
         let ctx = unsafe { &mut *ctx };
 
         let cert_cell = if let Some(cert_cell)
@@ -146,6 +149,7 @@ impl RnpKey {
     }
 
     fn find_cert(&mut self) {
+        rnp_function!(RnpKey::find_cert, crate::TRACE);
         if self.cert.is_none() {
             // There may be multiple certificates with the same key.
             // Prefer an exact match (i.e., fingerprint *and* secret
@@ -281,6 +285,7 @@ fn rnp_key_handle_destroy(key: *mut RnpKey)
 fn rnp_key_get_keyid(key: *const RnpKey,
                      keyid: *mut *mut c_char)
                      -> RnpResult {
+    rnp_function!(rnp_key_get_keyid, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(keyid);
 
@@ -292,6 +297,7 @@ fn rnp_key_get_keyid(key: *const RnpKey,
 fn rnp_key_get_fprint(key: *const RnpKey,
                      fprint: *mut *mut c_char)
                      -> RnpResult {
+    rnp_function!(rnp_key_get_fprint, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(fprint);
 
@@ -303,6 +309,7 @@ fn rnp_key_get_fprint(key: *const RnpKey,
 fn rnp_key_get_alg(key: *const RnpKey,
                    alg: *mut *mut c_char)
                    -> RnpResult {
+    rnp_function!(rnp_key_get_alg, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(alg);
 
@@ -310,6 +317,39 @@ fn rnp_key_get_alg(key: *const RnpKey,
     RNP_SUCCESS
 }
 
+#[no_mangle] pub unsafe extern "C"
+fn rnp_key_get_curve(key: *mut RnpKey, curve_out: *mut *mut c_char)
+                     -> RnpResult {
+    rnp_function!(rnp_key_get_curve, crate::TRACE);
+    let key = assert_ptr_ref!(key);
+    assert_ptr!(curve_out);
+
+    use openpgp::crypto::mpi::PublicKey;
+    match key.key.mpis() {
+        | PublicKey::EdDSA { curve, .. }
+        | PublicKey::ECDSA { curve, .. }
+        | PublicKey::ECDH { curve, .. } => {
+            use openpgp::types::Curve::*;
+            // We want to return an error on unknown curves.
+            match curve {
+                | NistP256
+                | NistP384
+                | NistP521
+                | BrainpoolP256
+                | BrainpoolP512
+                | Ed25519
+                | Cv25519 => (),
+                _ => return RNP_ERROR_BAD_PARAMETERS,
+            };
+
+            *curve_out = str_to_rnp_buffer(curve.to_rnp_id());
+            RNP_SUCCESS
+        },
+        // Not an ECC algorithm.
+        _ => RNP_ERROR_BAD_PARAMETERS
+    }
+}
+
 impl RnpKey {
     fn is_primary(&mut self) -> Result<bool> {
         let key_fpr = self.fingerprint();
@@ -328,6 +368,7 @@ impl RnpKey {
 fn rnp_key_is_primary(key: *mut RnpKey,
                       result: *mut bool)
                       -> RnpResult {
+    rnp_function!(rnp_key_is_primary, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(result);
     *result = rnp_try!((*key).is_primary());
@@ -341,6 +382,7 @@ fn rnp_key_is_primary(key: *mut RnpKey,
 fn rnp_key_is_sub(key: *mut RnpKey,
                   result: *mut bool)
                   -> RnpResult {
+    rnp_function!(rnp_key_is_sub, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(result);
     *result = ! rnp_try!((*key).is_primary());
@@ -351,6 +393,7 @@ fn rnp_key_is_sub(key: *mut RnpKey,
 fn rnp_key_get_grip(key: *const RnpKey,
                     grip: *mut *mut c_char)
                    -> RnpResult {
+    rnp_function!(rnp_key_get_grip, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(grip);
 
@@ -366,6 +409,7 @@ fn rnp_key_get_grip(key: *const RnpKey,
 fn rnp_key_get_primary_grip(key: *mut RnpKey,
                             grip: *mut *mut c_char)
                             -> RnpResult {
+    rnp_function!(rnp_key_get_primary_grip, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(grip);
 
@@ -387,6 +431,7 @@ fn rnp_key_get_primary_grip(key: *mut RnpKey,
 fn rnp_key_have_public(key: *const RnpKey,
                        result: *mut bool)
                        -> RnpResult {
+    rnp_function!(rnp_key_have_public, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(result);
 
@@ -398,6 +443,7 @@ fn rnp_key_have_public(key: *const RnpKey,
 fn rnp_key_have_secret(key: *const RnpKey,
                        result: *mut bool)
                        -> RnpResult {
+    rnp_function!(rnp_key_have_secret, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(result);
 
@@ -414,6 +460,7 @@ fn rnp_key_have_secret(key: *const RnpKey,
 fn rnp_key_get_bits(key: *const RnpKey,
                     bits: *mut u32)
                     -> RnpResult {
+    rnp_function!(rnp_key_get_bits, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(bits);
 
@@ -429,6 +476,7 @@ fn rnp_key_get_bits(key: *const RnpKey,
 fn rnp_key_get_creation(key: *const RnpKey,
                         creation: *mut u32)
                     -> RnpResult {
+    rnp_function!(rnp_key_get_creation, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(creation);
 
@@ -480,6 +528,7 @@ fn rnp_key_get_expiration(key: *mut RnpKey,
 fn rnp_key_get_primary_uid(key: *mut RnpKey,
                            uid: *mut *mut c_char)
                            -> RnpResult {
+    rnp_function!(rnp_key_get_primary_uid, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(uid);
 
@@ -511,6 +560,7 @@ fn rnp_key_get_uid_at(key: *mut RnpKey,
                       idx: size_t,
                       uid: *mut *mut c_char)
                       -> RnpResult {
+    rnp_function!(rnp_key_get_uid_at, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(uid);
 
@@ -536,6 +586,7 @@ fn rnp_key_get_uid_handle_at(key: *mut RnpKey,
                              idx: size_t,
                              uid: *mut *mut RnpUserID)
                              -> RnpResult {
+    rnp_function!(rnp_key_get_uid_handle_at, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(uid);
 
@@ -562,6 +613,7 @@ fn rnp_key_get_uid_handle_at(key: *mut RnpKey,
 fn rnp_key_get_uid_count(key: *mut RnpKey,
                          count: *mut size_t)
                          -> RnpResult {
+    rnp_function!(rnp_key_get_uid_count, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(count);
 
@@ -579,6 +631,7 @@ fn rnp_key_get_subkey_at(key: *mut RnpKey,
                          idx: size_t,
                          subkey: *mut *mut RnpKey)
                          -> RnpResult {
+    rnp_function!(rnp_key_get_subkey_at, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(subkey);
 
@@ -607,6 +660,7 @@ fn rnp_key_get_subkey_at(key: *mut RnpKey,
 fn rnp_key_get_subkey_count(key: *mut RnpKey,
                             count: *mut size_t)
                             -> RnpResult {
+    rnp_function!(rnp_key_get_subkey_count, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(count);
 
@@ -623,6 +677,7 @@ fn rnp_key_allows_usage(key: *mut RnpKey,
                         usage: *const c_char,
                         result: *mut bool)
                         -> RnpResult {
+    rnp_function!(rnp_key_allows_usage, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(usage);
     assert_ptr!(result);
@@ -654,6 +709,7 @@ fn rnp_key_allows_usage(key: *mut RnpKey,
 fn rnp_key_is_revoked(key: *mut RnpKey,
                       result: *mut bool)
                       -> RnpResult {
+    rnp_function!(rnp_key_is_revoked, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(result);
 
@@ -689,6 +745,7 @@ fn rnp_key_export_autocrypt(key: *mut RnpKey,
                             output: *mut RnpOutput,
                             _flags: u32)
                             -> RnpResult {
+    rnp_function!(rnp_key_export_autocrypt, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(uid);
     assert_ptr!(output);
@@ -729,6 +786,7 @@ fn rnp_key_export_autocrypt(key: *mut RnpKey,
 fn rnp_key_remove(key: *mut RnpKey,
                   flags: RnpKeyRemoveFlags)
                   -> RnpResult {
+    rnp_function!(rnp_key_remove, crate::TRACE);
     assert_ptr!(key);
     let key = &mut *key;
     let ctx = &mut *key.ctx;
@@ -775,6 +833,7 @@ fn rnp_key_export(key: *mut RnpKey,
                   output: *mut RnpOutput,
                   flags: RnpKeyExportFlags)
                   -> RnpResult {
+    rnp_function!(rnp_key_export, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(output);
     let key = &mut *key;
@@ -816,18 +875,7 @@ fn rnp_key_export(key: *mut RnpKey,
             (false, false) => cert.export(sink)?,
             (false,  true) => cert.as_tsk().export(sink)?,
             ( true, false) => cert.armored().export(sink)?,
-            ( true,  true) => {
-                // XXX: This is not available in sequoia-openpgp 1.0:
-                //
-                // cert.as_tsk().armored().export(sink)?
-                //
-                // Drop this workaround once we depend on >= 1.1, and
-                // use the above instead.
-                use openpgp::armor::{Writer, Kind};
-                let mut w = Writer::new(sink, Kind::SecretKey)?;
-                cert.as_tsk().serialize(&mut w)?;
-                w.finalize()?;
-            },
+            ( true,  true) => cert.as_tsk().armored().export(sink)?,
         }
         Ok(())
     }
@@ -844,6 +892,7 @@ fn rnp_key_export(key: *mut RnpKey,
 fn rnp_key_is_locked(key: *mut RnpKey,
                      result: *mut bool)
                      -> RnpResult {
+    rnp_function!(rnp_key_is_locked, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(result);
     let key = &mut *key;
@@ -858,6 +907,7 @@ fn rnp_key_is_locked(key: *mut RnpKey,
 #[no_mangle] pub unsafe extern "C"
 fn rnp_key_lock(key: *mut RnpKey)
                 -> RnpResult {
+    rnp_function!(rnp_key_lock, crate::TRACE);
     assert_ptr!(key);
     let key = &mut *key;
     if let Some(k) = key.key.parts_as_secret().ok() {
@@ -872,6 +922,7 @@ fn rnp_key_lock(key: *mut RnpKey)
 fn rnp_key_unlock(key: *mut RnpKey,
                   password: *const c_char)
                   -> RnpResult {
+    rnp_function!(rnp_key_unlock, crate::TRACE);
     assert_ptr!(key);
     let password = if password.is_null() {
         None
@@ -1087,6 +1138,7 @@ fn rnp_key_export_revocation(key: *mut RnpKey,
                              code: *const c_char,
                              reason: *const c_char)
                              -> RnpResult {
+    rnp_function!(rnp_key_export_revocation, crate::TRACE);
     assert_ptr!(key);
     assert_ptr!(output);
     let key = &mut *key;
@@ -1245,7 +1297,7 @@ fn rnp_key_revoke(key: *mut RnpKey,
     }
 
     if let Err(e) = f(ctx, cert, key, hash, code, reason) {
-        _warn!("{}", e);
+        warn!("{}", e);
         RNP_ERROR_GENERIC
     } else {
         RNP_SUCCESS
@@ -1294,7 +1346,7 @@ fn rnp_key_set_expiration(key: *mut RnpKey,
                     .into_keypair()?
             },
             Err(e) => {
-                _warn!("Cannot change expiration time without \
+                warn!("Cannot change expiration time without \
                         primary secret: {}", e);
                 return Err(e);
             },
@@ -1326,7 +1378,7 @@ fn rnp_key_set_expiration(key: *mut RnpKey,
                                  .into_keypair()?)
                         },
                         Err(e) => {
-                            _warn!("Cannot change expiration time without \
+                            warn!("Cannot change expiration time without \
                                     subkey secret: {}", e);
                             return Err(e);
                         },
@@ -1350,7 +1402,7 @@ fn rnp_key_set_expiration(key: *mut RnpKey,
     }
 
     if let Err(e) = f(ctx, cert, key, expiry) {
-        _warn!("{}", e);
+        warn!("{}", e);
         RNP_ERROR_GENERIC
     } else {
         RNP_SUCCESS
@@ -1362,6 +1414,21 @@ fn rnp_key_valid_till(key: *mut RnpKey,
                       result: *mut u32)
                       -> RnpResult {
     rnp_function!(rnp_key_valid_till, crate::TRACE);
+    let result = assert_ptr_mut!(result);
+
+    // The key expiration time is a 33-bit number, but the
+    // RNP API only has place for a 32-bit number :/.
+    let mut till = 0;
+    let r = rnp_key_valid_till64(key, &mut till as *mut _);
+    *result = min(u32::MAX as u64, till) as u32;
+    r
+}
+
+#[no_mangle] pub unsafe extern "C"
+fn rnp_key_valid_till64(key: *mut RnpKey,
+                        result: *mut u64)
+                        -> RnpResult {
+    rnp_function!(rnp_key_valid_till64, crate::TRACE);
     let key = assert_ptr_mut!(key);
     assert_ptr!(result);
 
@@ -1430,13 +1497,11 @@ fn rnp_key_valid_till(key: *mut RnpKey,
         });
     };
 
-    // The key expiration time is a 33-bit number, but the
-    // RNP API only has place for a 32-bit number :/.
     if let Ok(unix) = t
         .unwrap_or(std::time::UNIX_EPOCH + Duration::new(u32::MAX as u64, 0))
         .duration_since(std::time::UNIX_EPOCH)
     {
-        *result = min(u32::MAX as u64, unix.as_secs()) as u32;
+        *result = unix.as_secs();
         t!("valid till: {}", *result);
     } else {
         // Ummm... this shouldn't happen, but if it does
@@ -1447,6 +1512,248 @@ fn rnp_key_valid_till(key: *mut RnpKey,
     RNP_SUCCESS
 }
 
+#[no_mangle] pub unsafe extern "C"
+fn rnp_key_25519_bits_tweaked(key: *mut RnpKey, result_out: *mut bool)
+                              -> RnpResult {
+    rnp_function!(rnp_key_25519_bits_tweaked, crate::TRACE);
+    let key = assert_ptr_ref!(key);
+    let ctx = &mut *key.ctx;
+    assert_ptr!(result_out);
+
+    // First, check that we are a Curve25519 key.
+    use openpgp::crypto::mpi::PublicKey;
+    match key.key.mpis() {
+        | PublicKey::ECDH { curve, .. } => {
+            use openpgp::types::Curve::*;
+            // We want to return an error on unknown curves.
+            match curve {
+                Cv25519 => (),
+                _ => return RNP_ERROR_BAD_PARAMETERS,
+            };
+        },
+        // Not an ECC algorithm.
+        _ => return RNP_ERROR_BAD_PARAMETERS,
+    }
+
+    // RNP's implementation also considers unlocked keys.  Do the
+    // same.
+    let maybe_unlocked_key = ctx.key_unlocked_ref(&key.key)
+        .map(|k| k.parts_as_unspecified())
+        .unwrap_or(&key.key);
+
+    // Now check the secret.
+    use openpgp::packet::key;
+    match maybe_unlocked_key.optional_secret() {
+        Some(key::SecretKeyMaterial::Unencrypted(k)) => k.map(|secret| {
+            use openpgp::crypto::mpi::SecretKeyMaterial;
+            match secret {
+                SecretKeyMaterial::ECDH { scalar } => {
+                    // OpenPGP stores the secret in reverse order.
+                    const CURVE25519_SIZE: usize = 32;
+                    const FIRST: usize = CURVE25519_SIZE - 1;
+                    const LAST: usize = 0;
+
+                    // Curve25519 Paper, Sec. 3: A user can, for
+                    // example, generate 32 uniform random bytes,
+                    // clear bits 0, 1, 2 of the first byte, clear bit
+                    // 7 of the last byte, and set bit 6 of the last
+                    // byte.
+                    let scalar = scalar.value_padded(CURVE25519_SIZE);
+                    *result_out =
+                        scalar[FIRST] & !0b1111_1000 == 0
+                        && scalar[LAST] & 0b1000_0000 == 0
+                        && scalar[LAST] & 0b0100_0000 == 0b0100_0000;
+                    RNP_SUCCESS
+
+                },
+                _ => RNP_ERROR_BAD_PARAMETERS,
+            }
+        }),
+        _ => RNP_ERROR_BAD_PARAMETERS,
+    }
+}
+
+#[no_mangle] pub unsafe extern "C"
+fn rnp_key_25519_bits_tweak(key: *mut RnpKey)
+                            -> RnpResult {
+    rnp_function!(rnp_key_25519_bits_tweak, crate::TRACE);
+    let key = assert_ptr_mut!(key);
+
+    // First, check that we are a Curve25519 key.
+    use openpgp::crypto::mpi::PublicKey;
+    match key.key.mpis() {
+        | PublicKey::ECDH { curve, .. } => {
+            use openpgp::types::Curve::*;
+            // We want to return an error on unknown curves.
+            match curve {
+                Cv25519 => (),
+                _ => return RNP_ERROR_BAD_PARAMETERS,
+            };
+        },
+        // Not an ECC algorithm.
+        _ => return RNP_ERROR_BAD_PARAMETERS,
+    }
+
+    // Now check and clone the secret.
+    const CURVE25519_SIZE: usize = 32;
+    use openpgp::packet::key;
+    use openpgp::crypto::mpi;
+    let mut secret =
+        match key.key.optional_secret() {
+            Some(key::SecretKeyMaterial::Unencrypted(k)) =>
+                rnp_try!(k.map(|secret| {
+                    match secret {
+                        mpi::SecretKeyMaterial::ECDH { scalar } =>
+                            Ok(scalar.value_padded(CURVE25519_SIZE)),
+                        _ => Err(RNP_ERROR_BAD_PARAMETERS),
+                    }
+                })),
+            _ => return RNP_ERROR_BAD_PARAMETERS,
+        };
+
+    // OpenPGP stores the secret in reverse order.
+    const FIRST: usize = CURVE25519_SIZE - 1;
+    const LAST: usize = 0;
+
+    // Curve25519 Paper, Sec. 3: A user can, for example, generate 32
+    // uniform random bytes, clear bits 0, 1, 2 of the first byte,
+    // clear bit 7 of the last byte, and set bit 6 of the last byte.
+    secret[FIRST] &= 0b1111_1000;
+    secret[LAST] &= !0b1000_0000;
+    secret[LAST] |= 0b0100_0000;
+
+    // Finally, replace the secret in the key.
+    key.key = key.key.clone().add_secret(
+        key::SecretKeyMaterial::Unencrypted(
+            mpi::SecretKeyMaterial::ECDH { scalar: secret.into(), }
+            .into()))
+        .0.parts_into_unspecified();
+
+    // And update the cert.
+    let k = key.key.clone().parts_into_secret().unwrap();
+    if let Some(mut cert) = key.cert_mut() {
+        let key: openpgp::Packet =
+            if k.fingerprint() == cert.fingerprint() {
+                k.role_into_primary().into()
+            } else {
+                k.role_into_subordinate().into()
+            };
+
+        *cert = rnp_try_or!(cert.clone().insert_packets(Some(key)),
+                            RNP_ERROR_GENERIC);
+        RNP_SUCCESS
+    } else {
+        warn!("Key for {:X} not found in keyring", k.fingerprint());
+        RNP_ERROR_GENERIC
+    }
+}
+
+/// Removes the given UID from KEY.
+#[no_mangle] pub unsafe extern "C"
+fn rnp_uid_remove(key: *mut RnpKey, uid: *mut RnpUserID)
+                  -> RnpResult {
+    rnp_function!(rnp_uid_remove, crate::TRACE);
+    let key = assert_ptr_mut!(key);
+    let uid = assert_ptr_ref!(uid);
+
+    if let Some(mut cert) = key.cert_mut() {
+        *cert = cert.clone().retain_userids(|u| u.userid() != uid.userid());
+        RNP_SUCCESS
+    } else {
+        RNP_ERROR_GENERIC
+    }
+}
+
+type RnpKeySignaturesCB = fn(*mut RnpContext,
+                             *mut c_void,
+                             *mut RnpSignature,
+                             *mut RnpKeyRemoveSignatureAction);
+
+/// Removes signatures from KEY.
+#[no_mangle] pub unsafe extern "C"
+fn rnp_key_remove_signatures(key: *mut RnpKey,
+                             flags: RnpKeyRemoveSignatureFlags,
+                             pred: *mut RnpKeySignaturesCB,
+                             _cookie: *mut c_void)
+                             -> RnpResult {
+    rnp_function!(rnp_key_remove_signatures, crate::TRACE);
+    let key = assert_ptr_mut!(key);
+
+    let remove_invalid = flags & RNP_KEY_SIGNATURE_INVALID > 0;
+    if remove_invalid {
+        warn!("Flag RNP_KEY_SIGNATURE_INVALID not implemented, ignoring.");
+    }
+    let remove_unknown_key = flags & RNP_KEY_SIGNATURE_UNKNOWN_KEY > 0;
+    if remove_unknown_key {
+        warn!("Flag RNP_KEY_SIGNATURE_UNKNOWN_KEY not implemented, ignoring.");
+    }
+    let remove_non_selfsig = flags & RNP_KEY_SIGNATURE_NON_SELF_SIG > 0;
+
+    if ! pred.is_null() {
+        warn!("Predicate parameter not implemented, ignoring.");
+    }
+
+    if let Some(mut cert) = key.cert_mut() {
+        let mut acc = Vec::new();
+
+        // The primary key and related signatures.
+        let pk_bundle = cert.primary_key().bundle();
+        acc.push(pk_bundle.key().clone().into());
+        for s in pk_bundle.self_signatures() { acc.push(s.clone().into()); }
+        for s in pk_bundle.self_revocations()  { acc.push(s.clone().into()); }
+        if ! remove_non_selfsig {
+            for s in pk_bundle.other_revocations() { acc.push(s.clone().into()); }
+            for s in pk_bundle.certifications() { acc.push(s.clone().into()); }
+        }
+
+        // The subkeys and related signatures.
+        for skb in cert.keys().subkeys() {
+            acc.push(skb.key().clone().into());
+            for s in skb.self_signatures()   { acc.push(s.clone().into()); }
+            for s in skb.self_revocations()  { acc.push(s.clone().into()); }
+            if ! remove_non_selfsig {
+                for s in skb.other_revocations() { acc.push(s.clone().into()); }
+                for s in skb.certifications() { acc.push(s.clone().into()); }
+            }
+        }
+
+        // The UserIDs.
+        for uidb in cert.userids() {
+            acc.push(uidb.userid().clone().into());
+            for s in uidb.self_signatures()   { acc.push(s.clone().into()); }
+            for s in uidb.self_revocations()  { acc.push(s.clone().into()); }
+            for s in uidb.attestations()      { acc.push(s.clone().into()); }
+            if ! remove_non_selfsig {
+                for s in uidb.other_revocations() { acc.push(s.clone().into()); }
+                for s in uidb.certifications() { acc.push(s.clone().into()); }
+            }
+        }
+
+        // The UserAttributes.
+        for uab in cert.user_attributes() {
+            acc.push(uab.user_attribute().clone().into());
+            for s in uab.self_signatures()   { acc.push(s.clone().into()); }
+            for s in uab.self_revocations()  { acc.push(s.clone().into()); }
+            for s in uab.attestations()      { acc.push(s.clone().into()); }
+            if ! remove_non_selfsig {
+                for s in uab.other_revocations() { acc.push(s.clone().into()); }
+                for s in uab.certifications() { acc.push(s.clone().into()); }
+            }
+        }
+
+        if ! remove_invalid {
+            for s in cert.bad_signatures() { acc.push(s.clone().into()); }
+        }
+
+        *cert = rnp_try_or!(Cert::from_packets(acc.into_iter()),
+                            RNP_ERROR_GENERIC);
+        RNP_SUCCESS
+    } else {
+        RNP_ERROR_GENERIC
+    }
+}
+
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -1671,4 +1978,167 @@ mod tests {
 
         Ok(())
     }
+
+    #[test]
+    fn curve25519_secret_tweaking() {
+        fn test(buffer: &[u8]) {
+            let mut ctx: *mut RnpContext = std::ptr::null_mut();
+            let r = unsafe {
+                rnp_ffi_create(
+                    &mut ctx as *mut *mut _,
+                    b"GPG\x00".as_ptr() as *const c_char,
+                    b"GPG\x00".as_ptr() as *const c_char)
+            };
+            assert_eq!(r, RNP_SUCCESS);
+
+            let mut input: *mut RnpInput = std::ptr::null_mut();
+            let r = unsafe {
+                rnp_input_from_memory(
+                    &mut input as *mut *mut _,
+                    buffer.as_ptr(), buffer.len(), false)
+            };
+            assert_eq!(r, RNP_SUCCESS);
+
+            // And load the certificate.
+            let r = unsafe {
+                rnp_import_keys(
+                    ctx, input,
+                    RNP_LOAD_SAVE_PUBLIC_KEYS
+                        | RNP_LOAD_SAVE_SECRET_KEYS,
+                    std::ptr::null_mut() as *mut *mut _)
+            };
+            assert_eq!(r, RNP_SUCCESS);
+
+            let r = unsafe {
+                rnp_input_destroy(input)
+            };
+            assert_eq!(r, RNP_SUCCESS);
+
+            log!("ctx: keystore: {:?}", unsafe { &mut *ctx }.certs);
+
+            let mut key: *mut RnpKey = std::ptr::null_mut();
+            let r = unsafe {
+                rnp_locate_key(
+                    ctx,
+                    b"fingerprint\x00".as_ptr()
+                        as *const c_char,
+                    b"EA02B24FFD4C1B96616D3DF24766F6B9D5F21EB6\x00".as_ptr()
+                        as *const c_char,
+                    &mut key as *mut *mut _)
+            };
+            assert_eq!(r, RNP_SUCCESS);
+
+            let mut result = false;
+            let r = unsafe {
+                rnp_key_25519_bits_tweaked(key, &mut result)
+            };
+            assert_eq!(r, RNP_SUCCESS);
+            assert_eq!(result, false);
+
+            let r = unsafe {
+                rnp_key_25519_bits_tweak(key)
+            };
+            assert_eq!(r, RNP_SUCCESS);
+
+            // Check modification against reference.
+            let reference = b"-----BEGIN PGP PRIVATE KEY BLOCK-----
+Comment: EB85 BB5F A33A 75E1 5E94  4E63 F231 550C 4F47 E38E
+Comment: Alice Lovelace <alice@openpgp.example>
+
+xVgEXEcE6RYJKwYBBAHaRw8BAQdArjWwk3FAqyiFbFBKT4TzXcVBqPTB3gmzlC/U
+b7O1u10AAP9XBeW6lzGOLx7zHH9AsUDUTb2pggYGMzd0P3ulJ2AfvQ4RzSZBbGlj
+ZSBMb3ZlbGFjZSA8YWxpY2VAb3BlbnBncC5leGFtcGxlPsKQBBMWCAA4AhsDBQsJ
+CAcCBhUKCQgLAgQWAgMBAh4BAheAFiEE64W7X6M6deFelE5j8jFVDE9H444FAl2l
+nzoACgkQ8jFVDE9H447pKwD6A5xwUqIDprBzrHfahrImaYEZzncqb25vkLV2arYf
+a78A/R3AwtLQvjxwLDuzk4dUtUwvUYibL2sAHwj2kGaHnfICx10EXEcE6RIKKwYB
+BAGXVQEFAQEHQEL/BiGtq0k84Km1wqQw2DIikVYrQrMttN8d7BPfnr4iAwEIBwAA
+/3/xFPG6U17rhTuq+07gmEvaFYKfxRB6sgAYiW6TMTpQEK7CeAQYFggAIBYhBOuF
+u1+jOnXhXpROY/IxVQxPR+OOBQJcRwTpAhsMAAoJEPIxVQxPR+OOWdABAMUdSzpM
+hzGs1O0RkWNQWbUzQ8nUOeD9wNbjE3zR+yfRAQDbYqvtWQKN4AQLTxVJN5X5AWyb
+Pnn+We1aTBhaGa86AQ==
+=3GfK
+-----END PGP PRIVATE KEY BLOCK-----
+";
+            use openpgp::parse::Parse;
+            let reference_cert =
+                openpgp::Cert::from_bytes(reference).unwrap();
+            unsafe {
+                use std::ops::Deref;
+                assert_eq!(
+                    (*key).cert.as_ref().unwrap().as_ref()
+                    .read().unwrap().deref(),
+                    &reference_cert);
+            }
+
+            let mut result = false;
+            let r = unsafe {
+                rnp_key_25519_bits_tweaked(key, &mut result)
+            };
+            assert_eq!(r, RNP_SUCCESS);
+            assert_eq!(result, true);
+
+            let r = unsafe {
+                rnp_key_handle_destroy(key)
+            };
+            assert_eq!(r, RNP_SUCCESS);
+
+            let r = unsafe {
+                rnp_ffi_destroy(ctx)
+            };
+            assert_eq!(r, RNP_SUCCESS);
+        }
+
+        // Secret with LSB set.
+        test(b"-----BEGIN PGP PRIVATE KEY BLOCK-----
+
+xVgEXEcE6RYJKwYBBAHaRw8BAQdArjWwk3FAqyiFbFBKT4TzXcVBqPTB3gmzlC/U
+b7O1u10AAP9XBeW6lzGOLx7zHH9AsUDUTb2pggYGMzd0P3ulJ2AfvQ4RzSZBbGlj
+ZSBMb3ZlbGFjZSA8YWxpY2VAb3BlbnBncC5leGFtcGxlPsKQBBMWCAA4AhsDBQsJ
+CAcCBhUKCQgLAgQWAgMBAh4BAheAFiEE64W7X6M6deFelE5j8jFVDE9H444FAl2l
+nzoACgkQ8jFVDE9H447pKwD6A5xwUqIDprBzrHfahrImaYEZzncqb25vkLV2arYf
+a78A/R3AwtLQvjxwLDuzk4dUtUwvUYibL2sAHwj2kGaHnfICx10EXEcE6RIKKwYB
+BAGXVQEFAQEHQEL/BiGtq0k84Km1wqQw2DIikVYrQrMttN8d7BPfnr4iAwEIBwAA
+/3/xFPG6U17rhTuq+07gmEvaFYKfxRB6sgAYiW6TMTpREK/CeAQYFggAIBYhBOuF
+u1+jOnXhXpROY/IxVQxPR+OOBQJcRwTpAhsMAAoJEPIxVQxPR+OOWdABAMUdSzpM
+hzGs1O0RkWNQWbUzQ8nUOeD9wNbjE3zR+yfRAQDbYqvtWQKN4AQLTxVJN5X5AWyb
+Pnn+We1aTBhaGa86AQ==
+=iQ01
+-----END PGP PRIVATE KEY BLOCK-----
+");
+        // Secret with MSB set.
+        test(b"-----BEGIN PGP PRIVATE KEY BLOCK-----
+
+xVgEXEcE6RYJKwYBBAHaRw8BAQdArjWwk3FAqyiFbFBKT4TzXcVBqPTB3gmzlC/U
+b7O1u10AAP9XBeW6lzGOLx7zHH9AsUDUTb2pggYGMzd0P3ulJ2AfvQ4RzSZBbGlj
+ZSBMb3ZlbGFjZSA8YWxpY2VAb3BlbnBncC5leGFtcGxlPsKQBBMWCAA4AhsDBQsJ
+CAcCBhUKCQgLAgQWAgMBAh4BAheAFiEE64W7X6M6deFelE5j8jFVDE9H444FAl2l
+nzoACgkQ8jFVDE9H447pKwD6A5xwUqIDprBzrHfahrImaYEZzncqb25vkLV2arYf
+a78A/R3AwtLQvjxwLDuzk4dUtUwvUYibL2sAHwj2kGaHnfICx10EXEcE6RIKKwYB
+BAGXVQEFAQEHQEL/BiGtq0k84Km1wqQw2DIikVYrQrMttN8d7BPfnr4iAwEIBwAB
+AP/xFPG6U17rhTuq+07gmEvaFYKfxRB6sgAYiW6TMTpQEDDCeAQYFggAIBYhBOuF
+u1+jOnXhXpROY/IxVQxPR+OOBQJcRwTpAhsMAAoJEPIxVQxPR+OOWdABAMUdSzpM
+hzGs1O0RkWNQWbUzQ8nUOeD9wNbjE3zR+yfRAQDbYqvtWQKN4AQLTxVJN5X5AWyb
+Pnn+We1aTBhaGa86AQ==
+=YOA6
+-----END PGP PRIVATE KEY BLOCK-----
+");
+
+        // Secret with 2^254 bit cleared.
+        test(b"-----BEGIN PGP PRIVATE KEY BLOCK-----
+
+xVgEXEcE6RYJKwYBBAHaRw8BAQdArjWwk3FAqyiFbFBKT4TzXcVBqPTB3gmzlC/U
+b7O1u10AAP9XBeW6lzGOLx7zHH9AsUDUTb2pggYGMzd0P3ulJ2AfvQ4RzSZBbGlj
+ZSBMb3ZlbGFjZSA8YWxpY2VAb3BlbnBncC5leGFtcGxlPsKQBBMWCAA4AhsDBQsJ
+CAcCBhUKCQgLAgQWAgMBAh4BAheAFiEE64W7X6M6deFelE5j8jFVDE9H444FAl2l
+nzoACgkQ8jFVDE9H447pKwD6A5xwUqIDprBzrHfahrImaYEZzncqb25vkLV2arYf
+a78A/R3AwtLQvjxwLDuzk4dUtUwvUYibL2sAHwj2kGaHnfICx10EXEcE6RIKKwYB
+BAGXVQEFAQEHQEL/BiGtq0k84Km1wqQw2DIikVYrQrMttN8d7BPfnr4iAwEIBwAA
+/j/xFPG6U17rhTuq+07gmEvaFYKfxRB6sgAYiW6TMTpQEG3CeAQYFggAIBYhBOuF
+u1+jOnXhXpROY/IxVQxPR+OOBQJcRwTpAhsMAAoJEPIxVQxPR+OOWdABAMUdSzpM
+hzGs1O0RkWNQWbUzQ8nUOeD9wNbjE3zR+yfRAQDbYqvtWQKN4AQLTxVJN5X5AWyb
+Pnn+We1aTBhaGa86AQ==
+=Txnk
+-----END PGP PRIVATE KEY BLOCK-----
+");
+    }
 }
