@@ -321,6 +321,8 @@ impl Keystore {
                                       import_secret_keys: bool)
         -> openpgp::Result<()>
     {
+        rnp_function!(Keystore::load_keyring_in_background, super::TRACE);
+
         if let Err(err) = self.gpg_thread_start() {
             warn!("Failed to start gpg thread: {}", err);
             KeystoreData::load_keyring(
@@ -996,27 +998,33 @@ impl KeystoreData {
 
                 let mut cert_ref = cert_cell.write().unwrap();
 
-                // XXX: When comparing the certificates we want to do:
-                //
-                //    certa.as_tsk() == certb.as_tsk()
-                //
-                // When https://gitlab.com/sequoia-pgp/sequoia/-/issues/701
-                // is closed, change the following code, which is more
-                // conservative than necessary.
-                let has_secret_key_material = cert_ref.is_tsk() || cert.is_tsk();
-
-                if !has_secret_key_material && *cert_ref == cert {
+                if cert_ref.as_tsk() == cert.as_tsk() {
                     // They are identical.  There is nothing to do.
                     t!("Update doesn't change certificate.  Short-circuiting.");
                     return;
                 }
 
-                // Merge.
+                // Merge.  We want to prefer secret keys in `cert`
+                // over the one in the store.  Cert::insert_packets
+                // replaces existing key material.  Hence, we insert
+                // only those packets that are either secret keys or
+                // public ones that are not in `cert_ref`.
+                let known_keys: HashSet<Fingerprint> =
+                    cert_ref.keys().map(|k| k.fingerprint()).collect();
                 let cert_update = cert_ref.clone()
-                    .insert_packets(cert.clone().into_packets())
+                    .insert_packets(cert.into_packets().filter(|p| match p {
+                        // Insert only unknown public keys.
+                        Packet::PublicKey(k) =>
+                            ! known_keys.contains(&k.fingerprint()),
+                        Packet::PublicSubkey(k) =>
+                            ! known_keys.contains(&k.fingerprint()),
+
+                        // Insert everything else, including secret keys.
+                        _ => true,
+                    }))
                     .expect("does not fail if the primary keys are equal");
 
-                if !has_secret_key_material && *cert_ref == cert {
+                if cert_ref.as_tsk() == cert_update.as_tsk() {
                     // They are identical.  There is nothing to do.
                     t!("Update doesn't add anything new.  Short-circuiting.");
                     return;
@@ -2162,4 +2170,32 @@ mod tests {
 
         Ok(())
     }
+
+    // Inserting the public key must not remove secrets.
+    #[test]
+    fn key_cert_secrets_preserved() -> openpgp::Result<()> {
+        let (key, _) =
+            openpgp::cert::CertBuilder::general_purpose(None, Some("Hanna")).generate()?;
+        let fp = key.fingerprint();
+        let a = key.clone();
+        let b = key.strip_secret_key_material();
+
+        // Test both ways.
+        for (first, second) in vec![(&a, &b), (&b, &a)] {
+            eprintln!("First has {}secret, second has {}secret.",
+                      if first.is_tsk() { "" } else { "no " },
+                      if second.is_tsk() { "" } else { "no " });
+            let ks = Keystore::default();
+            let mut ks = ks.write();
+            ks.insert(first.clone());
+            ks.insert(second.clone());
+
+            let k = ks.by_primary_fp(&fp).unwrap();
+            assert!(k.is_tsk());
+            assert!(k.primary_key().has_unencrypted_secret());
+            assert_eq!(k.keys().unencrypted_secret().count(),
+                       k.keys().count());
+        }
+        Ok(())
+    }
 }
