@@ -1,3 +1,5 @@
+#![doc = include_str!("../README.md")]
+
 use std::{
     collections::{
         HashMap,
@@ -30,7 +32,7 @@ use openpgp::{
     },
     packet::{
         Key,
-        key::{SecretParts, UnspecifiedRole},
+        key::{SecretParts, UnspecifiedParts, UnspecifiedRole},
         UserID,
     },
     policy::{
@@ -56,67 +58,67 @@ macro_rules! stub {
 
 #[allow(dead_code)]
 #[macro_use]
-mod error;
+pub mod error;
 use error::*;
 
 #[allow(dead_code)]
-mod stubs;
+pub mod stubs;
 
 use sequoia_ipc::Keygrip;
 
-mod keystore;
+pub mod keystore;
 use keystore::Keystore;
 
-mod buffer;
+pub mod buffer;
 use buffer::*;
 
 #[allow(dead_code)]
-mod flags;
+pub mod flags;
 use flags::*;
 #[allow(dead_code)]
-mod io;
+pub mod io;
 use io::*;
 #[allow(dead_code)]
-mod utils;
+pub mod utils;
 use utils::*;
 #[allow(dead_code)]
-mod conversions;
+pub mod conversions;
 use conversions::*;
 
-mod version;
+pub mod version;
 #[allow(dead_code)]
-mod op_verify;
+pub mod op_verify;
 #[allow(dead_code)]
-mod op_encrypt;
+pub mod op_encrypt;
 #[allow(dead_code)]
-mod op_sign;
-mod recombine;
+pub mod op_sign;
+pub mod recombine;
 #[allow(dead_code)]
-mod op_generate;
+pub mod op_generate;
 #[allow(dead_code)]
-mod signature;
+pub mod signature;
 use signature::RnpSignature;
 #[allow(dead_code)]
-mod key;
+pub mod key;
 use key::RnpKey;
 #[allow(dead_code)]
-mod iter;
+pub mod iter;
 #[allow(dead_code)]
-mod userid;
+pub mod userid;
 use userid::RnpUserID;
 #[allow(dead_code)]
-mod import;
+pub mod import;
 // The gpg module is copied from OpenPGP CA.  We don't want to modify
 // it.
 #[allow(dead_code)]
-mod gpg;
+pub mod gpg;
 
-mod tbprofile;
-mod wot;
+pub mod tbprofile;
+pub mod wot;
 #[cfg(feature="net")]
-mod parcimonie;
+pub mod parcimonie;
 #[cfg(not(feature="net"))]
-mod parcimonie_stub;
+pub mod parcimonie_stub;
 #[cfg(not(feature="net"))]
 use parcimonie_stub as parcimonie;
 
@@ -177,6 +179,7 @@ fn rnp_ffi_create(ctx: *mut *mut RnpContext,
                   sec_fmt: *const c_char)
                   -> RnpResult
 {
+    rnp_function!(rnp_ffi_create, crate::TRACE);
     assert_ptr!(ctx);
     assert_ptr!(pub_fmt);
     assert_ptr!(sec_fmt);
@@ -200,6 +203,7 @@ fn rnp_ffi_destroy(ctx: *mut RnpContext) -> RnpResult {
 
 #[no_mangle] pub unsafe extern "C"
 fn rnp_ffi_set_log_fd(ctx: *mut RnpContext, _fd: c_int) -> RnpResult {
+    rnp_function!(rnp_ffi_set_log_fd, crate::TRACE);
     assert_ptr!(ctx);
     RNP_SUCCESS
 }
@@ -209,6 +213,7 @@ fn rnp_ffi_set_pass_provider(ctx: *mut RnpContext,
                              cb: RnpPasswordCb,
                              cookie: *mut c_void)
                              -> RnpResult {
+    rnp_function!(rnp_ffi_set_pass_provider, crate::TRACE);
     assert_ptr!(ctx);
     (*ctx).password_cb = Some((cb, cookie));
     RNP_SUCCESS
@@ -504,6 +509,15 @@ impl RnpContext {
         self.unlocked_keys.insert(key.fingerprint(), key);
         Ok(())
     }
+
+    /// Returns a reference to the unlocked key in the cache, if it
+    /// exists.
+    pub fn key_unlocked_ref(&self,
+                            key: &Key<UnspecifiedParts, UnspecifiedRole>)
+                            -> Option<&Key<SecretParts, UnspecifiedRole>>
+    {
+        self.unlocked_keys.get(&key.fingerprint())
+    }
 }
 
 #[no_mangle] pub unsafe extern "C"
@@ -657,6 +671,7 @@ fn rnp_save_keys(ctx: *mut RnpContext,
 fn rnp_get_public_key_count(ctx: *mut RnpContext,
                             count: *mut size_t)
                             -> RnpResult {
+    rnp_function!(rnp_get_public_key_count, crate::TRACE);
     assert_ptr!(ctx);
 
     // Make sure the agent listing is up to date.
@@ -704,22 +719,3 @@ fn rnp_get_secret_key_count(ctx: *mut RnpContext,
     t!("-> {}", *count);
     RNP_SUCCESS
 }
-
-// These are bound in RNPlib.jsm but not actually used.
-macro_rules! unused {
-    ($s: ident) => {
-        #[no_mangle] pub extern "C"
-        fn $s() -> RnpResult {
-            warn!("previously unused function is used: {}", stringify!($s));
-            RNP_ERROR_NOT_IMPLEMENTED
-        }
-    };
-}
-
-unused!(rnp_guess_contents);
-unused!(rnp_decrypt);
-unused!(rnp_symenc_get_aead_alg);
-unused!(rnp_symenc_get_cipher);
-unused!(rnp_symenc_get_hash_alg);
-unused!(rnp_symenc_get_s2k_iterations);
-unused!(rnp_symenc_get_s2k_type);
