@@ -182,6 +182,7 @@ fn rnp_op_verify_create<'a>(op: *mut *mut RnpOpVerify<'a>,
                             output: *mut RnpOutput<'a>)
                             -> RnpResult
 {
+    rnp_function!(rnp_op_verify_create, crate::TRACE);
     assert_ptr!(op);
     assert_ptr!(ctx);
     assert_ptr!(input);
@@ -203,6 +204,7 @@ fn rnp_op_verify_detached_create<'a>(op: *mut *mut RnpOpVerify<'a>,
                                      signature: *mut RnpInput)
                                      -> RnpResult
 {
+    rnp_function!(rnp_op_verify_detached_create, crate::TRACE);
     assert_ptr!(op);
     assert_ptr!(ctx);
     assert_ptr!(input);
@@ -285,13 +287,13 @@ fn rnp_op_verify_execute(op: *mut RnpOpVerify) -> RnpResult {
         if op.result.encrypted_message()
             && ! op.result.decrypted_successfully()
         {
-            _warn!("failed to decrypt: {}", e);
+            warn!("failed to decrypt: {}", e);
             RNP_ERROR_DECRYPT_FAILED
         } else if let Some(status) = op.result.verification_error() {
-            _warn!("failed to verify sig: {}", e);
+            warn!("failed to verify sig: {}", e);
             status
         } else {
-            _warn!("{}", e);
+            warn!("{}", e);
             // We need to return an error that Thunderbird
             // understands.  It does not understand RNP_ERROR_GENERIC
             // in this context.  See:
@@ -336,6 +338,7 @@ fn rnp_op_verify_get_signature_count(op: *mut RnpOpVerify,
 fn rnp_op_verify_signature_get_handle(sig: *mut RnpOpVerifySignature,
                                       handle: *mut *mut RnpSignature)
                                       -> RnpResult {
+    rnp_function!(rnp_op_verify_signature_get_handle, crate::TRACE);
     assert_ptr!(sig);
     assert_ptr!(handle);
     let sig = &*sig;
@@ -349,16 +352,16 @@ fn rnp_op_verify_signature_get_handle(sig: *mut RnpOpVerifySignature,
 #[no_mangle] pub unsafe extern "C"
 fn rnp_op_verify_signature_get_status(sig: *mut RnpOpVerifySignature)
                                       -> RnpResult {
+    rnp_function!(rnp_op_verify_signature_get_status, crate::TRACE);
     assert_ptr!(sig);
     (*sig).status
 }
 
 #[no_mangle] pub unsafe extern "C"
 fn rnp_op_verify_signature_get_key(sig: *mut RnpOpVerifySignature,
-                                   // XXX: I think we want the signing
-                                   // key here.
                                    key: *mut *mut RnpKey)
                                    -> RnpResult {
+    rnp_function!(rnp_op_verify_signature_get_key, crate::TRACE);
     assert_ptr!(sig);
     assert_ptr!(key);
     *key = if let Some(key) = (*sig).key.as_ref().cloned() {
@@ -374,6 +377,7 @@ fn rnp_op_verify_signature_get_times(sig: *mut RnpOpVerifySignature,
                                      created: *mut u32,
                                      expires: *mut u32)
                                      -> RnpResult {
+    rnp_function!(rnp_op_verify_signature_get_times, crate::TRACE);
     use std::time::UNIX_EPOCH;
     assert_ptr!(sig);
     if ! created.is_null() {
@@ -466,6 +470,7 @@ fn rnp_op_verify_get_recipient_count(op: *mut RnpOpVerify,
 fn rnp_recipient_get_keyid(recipient: *const PKESK,
                            keyid: *mut *mut c_char)
                            -> RnpResult {
+    rnp_function!(rnp_recipient_get_keyid, crate::TRACE);
     assert_ptr!(recipient);
     assert_ptr!(keyid);
     *keyid = str_to_rnp_buffer(format!("{:X}", (*recipient).recipient()));
@@ -476,6 +481,7 @@ fn rnp_recipient_get_keyid(recipient: *const PKESK,
 fn rnp_recipient_get_alg(recipient: *const PKESK,
                          alg: *mut *mut c_char)
                          -> RnpResult {
+    rnp_function!(rnp_recipient_get_alg, crate::TRACE);
     assert_ptr!(recipient);
     assert_ptr!(alg);
     *alg = str_to_rnp_buffer((*recipient).pk_algo().to_rnp_id());
